#include "../sen-fb/sen-fb.h"
#include<msp430.h>
/*
* senior remote control
*
* sen-fb.c
*
*
* Author: GEKU
*
*   Version 4.0
*/

#define c36kHz  214

enum MOD {IDLE,CTRL,CLEAR, RECEIVE} mod;
union ALL
{
    unsigned short buttons;
    struct PORTS
    {
        unsigned char p1;
        unsigned char p2;
    }
    port;
};
/* global variables */
union ALL all;
short edges;
signed char timeout_cnt; // 1: 250ms .. 127: 31,75s
signed char block_cnt;
signed char clear_enable;
unsigned short fsize;
unsigned short fpos;
unsigned short duration;
unsigned short xduration;

/* function prototypes */
inline unsigned short stop_timer1_A3(void);
inline void start_timer1_A3(void);

/*****************************************************************************/

#pragma vector=TIMER0_A0_VECTOR
__interrupt void timerA0_interrupt(void)
/*- Timeout Timer */
{
    timeout_cnt--;

    if ((timeout_cnt)<=0)
    {
        TA0CCTL0&=~CCIE;  // CCR1 interrupt disable
        TA0CTL=MC_0;      // STOP timer A0

        _BIC_SR(LPM0_EXIT);
    }
}
/*****************************************************************************/

#pragma vector=PORT1_VECTOR
__interrupt void port1_interrupt(void)
/*- Port 1 interrupt service routine */
{
    all.port.p1|=P1IFG;                 // store pressed button
    P1IFG=0;

    if (mod==RECEIVE)
    {
        TA0CTL=MC_0;   // STOP timer A0
        TA1CTL=MC_0;   // STOP timer A1

        _BIC_SR(LPM0_EXIT);
    }
    else
    {
        /* set timeout */
        timeout_cnt=1;                      // 1/4s
        TA0CTL=TASSEL_2+MC_1+ID_3+TACLR;    // SMCLK/8, up-mode
        TA0CCR0=T250MS;                      // 250ms delay
        TA0CCTL0=CCIE;                      // CCR1 interrupt enabled

        _BIC_SR(LPM4_EXIT);
        _BIS_SR(LPM0_bits+GIE);
    }
}
/*****************************************************************************/

#pragma vector=PORT2_VECTOR
__interrupt void port2_interrupt(void)
/*- Port 2 interrupt service routine */
{
    if ((P2IFG&ir_in)&&(mod==RECEIVE)) // receiving IR signal?
    {
        duration=TA1R;  // save duration count
        P2IE&=~ir_in;   // disable IR receiver interrupt
        P2IFG&=~ir_in;  // clear IR input interrupt flag

        // (re)start timer A1 and count until IR input has changed
        TA1CTL=TASSEL_2+MC_1+ID_1+TACLR;    // SMCLK/1=> 2MHz=> 0,5us, up-mode
        TA1CCR0=0xfff0;                     // maximum duration count 32,76ms
        TA1CCTL0=CCIE;                      // CCR1 interrupt enabled

        if (edges<(fsize-1)) // maximum edges detected?
        {
            if (duration!=0) // not the first edges?
            {
                duration/=29;       // reduce solution to 36kHz (28us)

                if (P2IES&ir_in)    // space?
                {
                    duration+=2;        // rectify space +30us
                }

                if ((duration+=(xduration))>=FIRST_BIG_TIME) // >6,96ms means a silent period
                {
                    xduration=0;                                    // clear extended duration

                    if (block_cnt<=3)
                    {
                        P2OUT&=~ir_out;
                        {
                            unsigned char i=0xFF;

                            while (!(duration&0x8000))
                            {
                                duration<<=1;
                                i--;
                            }

                            FCTL1=FWKEY+WRT;                    // set WRT bit for write operation
                            FCTL3=FWKEY;                        // clear LOCK bit
                            *((unsigned char *)fpos+edges)=i;   // write edge2edge big duration to flash
                            FCTL1=FWKEY;                        // clear WRT bit
                            FCTL3=FWKEY+LOCK;                   // set LOCK bit
                            edges++;                                        // count edge
                            block_cnt++;
                        }
                    }
                    else
                    {
                        P2OUT^=ir_out;
                    }
                }
                else
                {
                    FCTL1=FWKEY+WRT;                            // set WRT bit for write operation
                    FCTL3=FWKEY;                                // clear LOCK bit
                    *((unsigned char *)fpos+edges)=duration;    // write edge2edge duration to flash
                    FCTL1=FWKEY;                                // clear WRT bit
                    FCTL3=FWKEY+LOCK;                           // set LOCK bit
                    edges++;                                        // count edge
                    block_cnt=0;
                }
            }

            // toggle level sensitivity of IR input (detect other edge)

            if (P2IES&ir_in)
            {
                P2IES&=~ir_in;
            }
            else
            {
                P2IES|=ir_in;
            }

            P2IFG&=~ir_in;  // clear IR input interrupt flag
            P2IE|=ir_in;    // enable IR input interrupt
        }
        else
        {
            TA0CTL=MC_0;        // STOP timeout
            TA1CTL=MC_0;        // STOP pulse duration measurement
            _BIC_SR(LPM0_EXIT); // start calculation immediately
        }
    }
    else // button2 pressed
    {
        all.port.p2|=(P2IFG&buttons_p2);    // store pressed port2 buttons
        P2IFG=0;                            // clear port2 interrupt flag

        if (mod==RECEIVE)           // pressed buttons released during IR receiving?
        {
            TA0CTL=MC_0;            // STOP timeout
            TA1CTL=MC_0;            // STOP pulse duration measurement
            _BIC_SR(LPM0_EXIT);     // start calculation immediately
        }
        else
        {
            /* set timeout */
            timeout_cnt=1;                      // 1/4s
            TA0CTL=TASSEL_2+MC_1+ID_3+TACLR;    // SMCLK/8, up-mode
            TA0CCR0=T250MS;                     // 250ms delay
            TA0CCTL0=CCIE;                      // CCR1 interrupt enabled

            _BIC_SR(LPM4_EXIT);
            _BIS_SR(LPM0_bits+GIE);             // wait for 2nd button pressed
        }
    }
}
/*****************************************************************************/

#pragma vector=TIMER1_A0_VECTOR
__interrupt void pulse_duration_overflow(void)
/*- pulse duration measurement overflow after 0xFFF0 = 1,9sec */
{
    if (xduration<(0xFFF0-2259))
    {
        xduration+=2259;                // 0xFFF0/29
    }

    TA1CTL=TASSEL_2+MC_1+ID_1+TACLR;    // SMCLK/1=> 2MHz=> 0,5us, up-mode
    TA1CCR0=0xfff0;                     // maximum duration count 32,76ms
    TA1CCTL0=CCIE;                      // CCR1 interrupt enabled
}
/*****************************************************************************/

inline void clear_button_code(unsigned short fpos, unsigned short fsize)
/* clearing storage for button code, depending on file size 256/512 */
{
    short int i;
    char *pFlash;                               // Flash pointer
    unsigned char temporary_irc_buffer [256];        // temporary IRC_BUFFER

    if ((fsize!=256)&&(fsize!=512)) return;     // wrong block size

    if ((fpos>0xFDFF)||(fpos<0xCE00)) return;   // wrong address range

    pFlash=(char *)fpos;                        // initialize flash pointer

    {
        unsigned char dirty_flag=0;

        for (i=(fsize-1);i>=0;i--)              // check erased flash
        {
            if (*(pFlash+i)!=0xff)
            {
                dirty_flag=1;
                break;
            }
        }

        if (!dirty_flag) return;                // nothing to clear?
    }

    if (fsize==256) // clear standard code storage (255 edges)
    {
        pFlash=(char *)(fpos^0x0100);

        /* save opposite 256 byte block*/
        for (i=(fsize-1);i>=0;--i)
        {
            temporary_irc_buffer [(unsigned short)i]=pFlash[(unsigned short)i];
        }

        /* clear 256 byte block */
        FCTL1=FWKEY+ERASE;          // set erase bit
        FCTL3=FWKEY;                // clear lock bit
        *pFlash=0;                  // dummy write to erase flash segment
        FCTL3=FWKEY+LOCK;           // set lock bit

        for (i=(fsize-1);i>=0;--i)  // check erased flash
        {
            temporary_irc_buffer [(unsigned short)i]=pFlash[(unsigned short)i];
        }

        /* restore opposite 256 byte block*/

        FCTL1=FWKEY+WRT;            // set WRT bit for write operation
        FCTL3=FWKEY;                // clear LOCK bit
        for (i=(fsize-1);i>=0;--i)  // check erased flash
        {
            pFlash[(unsigned short)i]=temporary_irc_buffer [(unsigned short)i];
        }
        FCTL1=FWKEY;                // clear WRT bit
        FCTL3=FWKEY+LOCK;           // set LOCK bit
    }
    else // clear double code storage (511 edges)
    {
        FCTL1=FWKEY+ERASE;          // set erase bit
        FCTL3=FWKEY;                // clear lock bit
        *(char *)fpos=0;            // dummy write to erase flash segment
        FCTL3=FWKEY+LOCK;           // set lock bit

    }
}
/*****************************************************************************/

inline void clear_all(void)
/* clearing button code storage at 0xCE00 .. 0xFC00 */
{
    char *pFlash;                           // flash pointer

    pFlash=(char *)0xCE00;

    while (pFlash<(char *)0xFE00)
    {
        FCTL1=FWKEY+ERASE;  // set erase bit
        FCTL3=FWKEY;        // clear lock bit
        *pFlash=0;          // dummy write to erase flash segment
        FCTL3=FWKEY+LOCK;   // set lock bit

        pFlash+=512;        // increment flash pointer
    }
}
/*****************************************************************************/

void init_ports(void)
{
    // initialize PORT 1
    P1SEL=0;
    P1DIR=0;
    P1REN=buttons_p1; // enable pull ups
    P1OUT=buttons_p1; // set pull ups
    P1IES=buttons_p1; // set interrupt sensitivity to high->low transition
    P1IFG=0;          // clear all Port 1 interrupts
    // initialize PORT 2
    P2SEL=0;
    P2DIR=ir_out;           // set ir_out to output
    P2REN=buttons_p2|ir_in; // enable pull ups
    P2OUT=buttons_p2|ir_in; // set pull ups
    P2IES=buttons_p2|ir_in; // set interrupt sensitivity to high->low transition
    P2IFG=0;                // clear all Port 2 interrupts
    P2OUT&=~ir_out;         // deactivate IR-LED
    // finish Initialization by enabling button interrupts
    P1IE=buttons_p1;  // enable interrupts buttons port1
    P2IE=buttons_p2;  // enable interrupts buttons port2
}
/*****************************************************************************/

inline void response(void)
{
    unsigned int j=0xffff;

    do
    {
        P2OUT|=ir_out;
        __delay_cycles(c36kHz);     // 13,4us
        P2OUT&=~ir_out;
        __delay_cycles(c36kHz);     // 13,4us
    }
    while (j--);                    // 1,75s
}
/*****************************************************************************/

inline void receive_ir(void)
{
    mod=RECEIVE;
    edges=0;
    TA1CTL=TASSEL_2+MC_0+ID_1+TACLR;    // STOP timer A1

    /* set timeout */
    timeout_cnt=40;                     // 250ms*40=10sec
    TA0CTL=TASSEL_2+MC_1+ID_3+TACLR;    // SMCLK/8, up-mode
    TA0CCR0=T250MS;                     // 250ms delay
    TA0CCTL0=CCIE;                      // CCR1 interrupt enabled

    duration=0;
    xduration=0;
    block_cnt=0;

    /* prepair port interrupts*/
    P1IE=0;
    P2IE=0;
    P1IES&=~buttons_p1;                 // set interrupt sensitivity to low->high transition
    P2IES&=~buttons_p2;                 // set interrupt sensitivity to low->high transition
    P2IES|=ir_in;                       // set interrupt sensitivity to high->low transition
    P1IFG=0;                            // clear all Port 1 interrupts
    P2IFG=0;                            // clear all Port 2 interrupts

    P1IE=buttons_p1;                    // enable button port 1 interrupts
    P2IE=buttons_p2|ir_in;              // enable button Port 2 and IR receiver interrupt

    _BIS_SR(LPM0_bits+GIE);             // wait for End of IR-Code
    /*---------------------------------------------------------------------------*/
    P1IE=0;                             // disable all port 1 interrupts
    P2IE=0;                             // disable all port 2 interrupts

    P1IFG=0;                            // clear all Port 1 interrupts
    P2IFG=0;                            // clear all Port 2 interrupts
    P2OUT&=~ir_out;                     // deactivate IR LED

    if ((edges)!=0)                     // any IR code (edges) received?
    {
        FCTL1=FWKEY+WRT;                            // set WRT bit for write operation
        FCTL3=FWKEY;                                // clear LOCK bit
        *((unsigned char *)fpos+edges)=0;    // mark end of record
        FCTL1=FWKEY;                                // clear WRT bit
        FCTL3=FWKEY+LOCK;                           // set LOCK bit

        response();
    }
    all.buttons=0;
    mod=IDLE;
}
/*****************************************************************************/

inline void transmit_ir(void)
{
    unsigned char *px;
    unsigned char cnt;
    unsigned int tx_duration;

    while (1)
    {
        cnt=0;
        px=(unsigned char *)fpos;

        while ((tx_duration=*px)!=0)
        {
            if ((tx_duration>=FIRST_BIG_TIME)&&(tx_duration<=LAST_BIG_TIME))
            {
                if (((((P1IN&buttons_p1)==buttons_p1)&&((P2IN&buttons_p2)==buttons_p2))&&(fsize==256))) break;

                tx_duration=1<<(tx_duration&0x0F);
            }

            P2OUT&=~ir_out;  //IR LED off
            cnt^=1;

            if (cnt)
            {
                while (tx_duration)
                {
                    P2OUT|=ir_out;
                    __delay_cycles(c36kHz);
                    P2OUT&=~ir_out;
                    __delay_cycles(c36kHz);
                    tx_duration--;
                }
            }
            else
            {
                while (tx_duration)
                {
                    P2OUT|=~ir_out;
                    __delay_cycles(c36kHz);
                    P2OUT&=~ir_out;
                    __delay_cycles(c36kHz);
                    tx_duration--;
                }
            }

            px++;
        }
        if (fsize==512) break;
        if ((((P1IN&buttons_p1)==buttons_p1)&&((P2IN&buttons_p2)==buttons_p2))&&(fsize==256)) break;

        P2OUT&=~ir_out;  //IR LED off
    }
    __delay_cycles(30);
    all.buttons=0;
}
/*****************************************************************************/

extern unsigned char code_table_dat[];

inline void decode_buttons(void)
{
    fpos=(unsigned short)code_table_dat; //0xCE00;

    switch (all.buttons)
    {
        /*---------------------------------*/
        case PROG1:     //  1: D000 - D1FF
            fsize=512;
            fpos+=0x200;
            break;
        case PROG2:     //  2: D200 - D3FF
            fsize=512;
            fpos+=0x400;
            break;
        case PROG3:     //  3: D400 - D5FF
            fsize=512;
            fpos+=0x600;
            break;
        case PROG4:     //  4: D600 - D7FF
            fsize=512;
            fpos+=0x800;
            break;
        case PROG5:     //  5: D800 - D9FF
            fsize=512;
            fpos+=0xA00;
            break;
        case PROG6:     //  6: DA00 - DBFF
            fsize=512;
            fpos+=0xC00;
            break;
        case PROG7:     //  7: DC00 - DDFF
            fsize=512;
            fpos+=0xE00;
            break;
        case PROG8:     //  8: DE00 - DFFF
            fsize=512;
            fpos+=0x1000;
            break;
        case PROG9:
            fsize=512;  //  9: E000 - E1FF
            fpos+=0x1200;
            break;
        case PROG10:    // 10: E200 - E3FF  (PROG1+PROG2)
            fsize=512;
            fpos+=0x1400;
            break;
        case PROG11:    // 11: E400 - E5FF  (PROG2+PROG3)
            fsize=512;
            fpos+=0x1600;
            break;
        case PROG12:    // 12: E600 - E7FF  (PROG3+PROG4)
            fsize=512;
            fpos+=0x1800;
            break;
        case PROG13:    // 13: E800 - E9FF  (PROG4+PROG5)
            fsize=512;
            fpos+=0x1A00;
            break;
        case PROG14:    // 14: EA00 - EBFF  (PROG5+PROG6)
            fsize=512;
            fpos+=0x1C00;
            break;
        case PROG15:    // 15: EC00 - EDFF  (PROG6+PROG7)
            fsize=512;
            fpos+=0x1E00;
            break;
        case PROG16:    // 16: EE00 - EFFF  (PROG7+PROG8)
            fsize=512;
            fpos+=0x2000;
            break;
        case PROG17:    // 17: F000 - F1FF  (PROG8+PROG9)
            fsize=512;
            fpos+=0x2200;
            break;
        /*---------------------------------*/
        case L_S:       // 18: F200 - F2FF
            fsize=256;
            fpos+=0x2400;
            break;
        case PROG_UP:   // 19: F300 - F3FF
            fsize=256;
            fpos+=0x2500;
            break;
        case PROG_DN:   // 20: F400 - F4FF
            fsize=256;
            fpos+=0x2600;
            break;
        case SOFTEN:    // 21: F500 - F5FF
            fsize=256;
            fpos+=0x2700;
            break;
        case LOUDER:    // 22: F600 - F6FF
            fsize=256;
            fpos+=0x2800;
            break;
        case ON_OFF:    // 23: F700 - F7FF
            fsize=256;
            fpos+=0x2900;
            break;
        /*---------------------------------*/
        case PU_PROG1:   // 24: F800 - F9FF
            fsize=512;
            fpos+=0x2A00;
            break;
        case PU_PROG2:   // 25: FA00 - FBFF
            fsize=512;
            fpos+=0x2C00;
            break;
        case PU_PROG3:   // 26: FC00 - FDFF
            fsize=512;
            fpos+=0x2E00;
            break;
        /*---------------------------------*/
        case CLEAR_PROG:    //  (PROG_UP+PROG_DN)
            if (clear_enable)
            {
                if (mod==CLEAR)
                {
                    clear_all();
                    mod=IDLE;
                    WDTCTL=WDTPW;
                    while(1){};
                }
                else
                {
                    mod=CLEAR;
                }
            }
            else // clear_disable
            {
                fsize=512;
                // fpos=0xCE00;
            }
            break;
        case CLEAR_ENABLE:  //  (PROG_UP+ON_OFF)
            clear_enable=1;
            mod=CTRL;
            response();
            break;
        case CLEAR_DISABLE: //  (PROG_DN+ON_OFF)
            clear_enable=0;
            mod=CTRL;
            response();
            break;
        /*---------------------------------*/
        default:
            mod=IDLE;
        }
}
/*****************************************************************************/

int main(void)
{
    WDTCTL=WDTPW|WDTHOLD;   // stop watchdog timer
    /*Set DCO to 16MHz */
    BCSCTL1=CALBC1_16MHZ;
    DCOCTL=CALDCO_16MHZ;
    BCSCTL2|=DIVS_3;        // Sub System Clock SMCLK DCO/8=2MHz
    FCTL2=0xa56c;           // FN=44 (0x2c)=> Fftg=MCLK/44=16 MHz/44=364kHz (needed for flash unit)
    init_ports();
    mod=IDLE;

    response(); // IR-led test

    _BIS_SR(LPM4_bits+GIE);

    while (1)
    {
        decode_buttons();

        if (mod==CTRL)
        {
            mod=IDLE;
        }
        else if (mod==CLEAR)
        {
            if (all.buttons!=CLEAR_PROG)
            {
                clear_button_code(fpos,fsize);
                response();
                mod=IDLE;
            }
        }
        else
        {
            if (*((unsigned char *)fpos)!=0xFF)
            {
                transmit_ir();  // send IR-code
            }
            else
            {
                receive_ir();   // learn new code
            }
        }

        all.buttons=0;
        TA0CTL=MC_0; // STOP timer A0
        TA1CTL=MC_0; // STOP timer A1
        init_ports();

        _BIS_SR(LPM4_bits+GIE);
    };
}
/*****************************************************************************/
